import 'dart:convert';

import 'package:irishman/irishman.dart';
import 'dart:io';

/// A class that stores data about a programming language. The data stored is
/// everything, from metadata, to token data, to AST information, etc.
class Language {
  String name = "";
  String description = "";
  String emoji = "";

  List<String> tokens = [];
  List<LanguageSymbol> symbols = [];

  /// The default constructor for a language class.
  Language(
    this.name, {
    this.tokens = const [],
    this.symbols = const [],
  });

  /// Converts a JSON file into a language class.
  factory Language.fromJSONFile(String path) {
    String fileData = File(path).readAsStringSync();
    return Language.fromJSONString(fileData);
  }

  /// Converts a JSON string into a language class.
  factory Language.fromJSONString(String jsonText) {
    Map<String, dynamic> json = jsonDecode(jsonText);

    return Language.fromJson(json);
  }

  /// Coverts JSON data into a language class.
  Language.fromJson(Map<String, dynamic> json) {
    name = json['name'] as String;
    description = json['description'] as String;
    emoji = json['emoji'] as String;

    if (json['tokens'] != null) {
      tokens = List.from(json['tokens']);
    }

    if (json['symbols'] != null) {
      List symbolList = json['symbols'] as List;
      symbols = symbolList
          .map(
            (e) => LanguageSymbol.fromJson(e),
          )
          .toList();
    }
  }

  /// Gets a tokenizer from a language definition.
  Tokenizer getTokenizer() {
    // If we don't presort keywords by length, they will falsely trigger
    // For example: `do` could trigger when it should be `doodle`
    List<String> sortedTokens = [...tokens];
    sortedTokens.sort((a, b) => b.length.compareTo(a.length));

    // Same goes for the symbols
    List<LanguageSymbol> sortedSymbols = [...symbols];
    sortedSymbols
        .sort((a, b) => b.activator.length.compareTo(a.activator.length));

    return Tokenizer(
      keywords: [
        ...sortedTokens.map((e) => Keyword(e)),
      ],
      symbols: [
        ...sortedSymbols.map((e) => Symbol(e.activator, e.nickname)),
      ],
      spans: [],
    );
  }
}

class LanguageSymbol {
  /// The name of the token.
  String nickname = "";

  /// A sequence of characters that activates a symbol token.
  String activator = "";

  /// The default constructor for a language symbol.
  LanguageSymbol(this.nickname, this.activator);

  /// Converts JSON data into a language symbol.
  LanguageSymbol.fromJson(Map<String, dynamic> json) {
    nickname = json['name'] as String;
    activator = json['activator'] as String;
  }
}
